import java.util.ArrayList;
import java.util.Objects;
import com.google.java.contract.*;

public class Lexer {


    @Requires({"s != null"})
    public ArrayList<Token> lex(String s) {

        ArrayList<String> separatedChars = separateChars(s);

        ArrayList<String> tokensAsStrings = buildSubstringsFromSeparatedChars(separatedChars);

        return constructTokenList(tokensAsStrings);



    }

    private ArrayList<String> separateChars(String stringToTokenize) {

        ArrayList<String> separated = new ArrayList<>();

        separated.add("|");

        for (char c : stringToTokenize.toCharArray()) {

            if(Character.isWhitespace(c)){
                continue;
            }

            if(Character.isDigit(c)) {
                separated.add( String.valueOf(c) );
            } else {
                separated.add("|");
                separated.add( String.valueOf(c) );
                separated.add("|");
            }
        }

        separated.add("|");

        return separated;
    }

    private ArrayList<String> buildSubstringsFromSeparatedChars(ArrayList<String> separated) {

        ArrayList<String> tokensAsStrings = new ArrayList<>();

        StringBuilder temp = new StringBuilder();

        for(String c : separated) {
            if(Objects.equals(c, "|")) {

                if(!temp.toString().equals("")) {
                    tokensAsStrings.add(temp.toString());
                }

                temp = new StringBuilder();

            } else {
                temp.append(c);
            }
        }

        return tokensAsStrings;

    }

    private ArrayList<Token> constructTokenList(ArrayList<String> tokensAsStrings) {

        ArrayList<Token> tokens = new ArrayList<>();

        boolean notANumber = false;

        for(String s1 : tokensAsStrings) {

            try {
                Integer.parseInt(s1);
                tokens.add(new Token(TokenType.number, s1));
            } catch (NumberFormatException ignored) {

                notANumber = true;

            }

            if(notANumber) {

                if (Character.isLetter( s1.charAt(0) )) {
                    tokens.add(new Token(TokenType.var, s1));
                } else {
                    tokens.add(new Token(TokenType.special, s1));
                }
                notANumber = false;
            }
        }

        return tokens;
    }
}
