import java.text.ParseException;
import java.util.ArrayList;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

public class Parser {

    public AstExpression parse (ArrayList<Token> tokenList) throws Exception {

        AstExpression root;

        root = parseAstExpression(tokenList);

        return root;
    }

    private AstExpression parseAstExpression (ArrayList<Token> tokenList) throws Exception {

        AstExpression toReturn = new AstExpression();

        // If tokenList contains one item: Expression is a value

        boolean isValue = false;

        if(tokenList.size() == 1 ) {

            toReturn.astValue = parseValue(tokenList);
            return toReturn;

        }

        // If tokenList contains a dot and not operators (+ - * /): Expression is a value
        for (Token t : tokenList) {


            if(t.getTokenString().equals(".")) {
                isValue = true;
            }


            if(t.getTokenString().equals("+") || t.getTokenString().equals("-") ||
                    t.getTokenString().equals("*") || t.getTokenString().equals("/") ||
                    t.getTokenString().equals("^")) {
                isValue = false;
                break;
            }

        }

        if(isValue) {
            toReturn.astValue = parseValue(tokenList);
            return toReturn;
        }


        // If method reaches this point: Expression must be expression in brackets or binary operation

        // Check if first and last tokens have "(" and ")" strings

        try {

            if (tokenList.get(0).tokenString.equals("(") && tokenList.get(tokenList.size() - 1).tokenString.equals(")")) {


                // Check if "false alarm", e.g. ( 1 + 2 ) + ( 1 + 2 ) <- starts and ends with ( and ) but is to be treated as binaryOp

                boolean falseAlarm = false;

                for (int i = tokenList.size() - 2; i > 0; i--) {

                    if (tokenList.get(i).tokenString.equals(")") || tokenList.get(i).tokenString.equals("(")) {

                        if (tokenList.get(i).tokenString.equals(")")) {
                            falseAlarm = false;

                        } else {
                            falseAlarm = true;
                        }
                        break;
                    }
                }

                // Parse tokenList without first and last object and save it in toReturns astExpression

                if (!falseAlarm) {

                    toReturn.astExpression = parseAstExpression(
                            (ArrayList<Token>) IntStream.range(1, tokenList.size() - 1)
                                    .mapToObj(i -> tokenList.get(i))
                                    .collect(Collectors.toList()) // https://www.baeldung.com/java-stream-indices
                    );
                    return toReturn;
                }

            }

        } catch (IndexOutOfBoundsException indexOutOfBoundsException) {
            throw new Exceptions.ParserException("String is not a valid function");
        }


        toReturn.astBinaryOp = parseBinaryOp(tokenList);

        return toReturn;

    }



    private AstBinaryOp parseBinaryOp (ArrayList<Token> tokenList) throws Exception {


        AstBinaryOp toReturn = new AstBinaryOp();

        int positionOfFirstTopLevelOperator = 0;
        int numberOfBrackets = 0;
        int i = 0;
        String lastBracket = ")";

        for(Token t : tokenList) {


            if(t.getTokenString().equals("(") || t.getTokenString().equals(")")) {
                numberOfBrackets++;
                lastBracket = t.tokenString;
            }

            if( ((t.getTokenString().equals("+") || t.getTokenString().equals("-") || t.getTokenString().equals("*")
                    || t.getTokenString().equals("/") || t.getTokenString().equals("^")) && (numberOfBrackets % 2 == 0) ) && lastBracket.equals((")") )) {

                positionOfFirstTopLevelOperator = i;
                break;

            }
            i++;

        }

        // Make two Token lists before and after the expression, parse each as expression

        toReturn.astExpression1 = parseAstExpression(new ArrayList<>(tokenList.subList(0, positionOfFirstTopLevelOperator)));
        toReturn.astOperator = parseOperator(tokenList.get(positionOfFirstTopLevelOperator));
        toReturn.astExpression2 = parseAstExpression(new ArrayList<>(tokenList.subList(positionOfFirstTopLevelOperator + 1, tokenList.size())));


        return toReturn;
    }

    private AstOperator parseOperator (Token operator) {

        AstOperator toReturn = new AstOperator();
        toReturn.astOperatorContent = operator.tokenString;

        return toReturn;
    }

    private AstValue parseValue (ArrayList<Token> valueTokens) {

        AstValue toReturn = new AstValue();

        boolean containsSpecialToken = false;

        for(Token t : valueTokens) {

            if(t.getTokenType().equals(TokenType.special)) {

                containsSpecialToken = true;

            }

        }

        if(containsSpecialToken) {

            toReturn.astDecimal = parseDecimal(valueTokens);

        } else {

            // Check if token is number or variable and parse accordingly

            if(valueTokens.get(0).getTokenType().equals(TokenType.var)) {

                toReturn.astVariable = parseVariable(valueTokens);

            } else {
                toReturn.astNumber = parseNumber(valueTokens);
            }


        }
        return toReturn;
    }

    private AstVariable parseVariable (ArrayList<Token> valueTokens) {
        AstVariable toReturn = new AstVariable();

        toReturn.astVariable = valueTokens.get(0).getTokenString();

        return toReturn;
    }

    private AstDecimal parseDecimal (ArrayList<Token> decimalTokens) {

        AstDecimal toReturn = new AstDecimal();

        int i = 0;
        int positionOfDecimalPoint = -1;

        for(Token t : decimalTokens) {

            if(t.getTokenType().equals(TokenType.special)) {
                positionOfDecimalPoint = i;
                break;
            }

            i++;

        }

        // parse digits before comma as list and save it into toReturn.before



        if (positionOfDecimalPoint == 1) {

            ArrayList<Token> temp = new ArrayList<>();
            temp.add(decimalTokens.get(0));
            toReturn.astDigitBeforeComma = parseNumber(temp);

        } else {


            try {

                toReturn.astDigitBeforeComma = parseNumber(new ArrayList<>(decimalTokens.subList(0, positionOfDecimalPoint - 1)));

            } catch (IllegalArgumentException illegalArgumentException) {
                throw new Exceptions.ParserException("String contains invalid char");
            }

        }


        if (positionOfDecimalPoint +  1 - decimalTokens.size() == 0) {

            ArrayList<Token> temp = new ArrayList<>();
            temp.add(decimalTokens.get(decimalTokens.size() - 1));
            toReturn.astDigitAfterComma = parseNumber(temp);

        } else {

            toReturn.astDigitAfterComma = parseNumber(new ArrayList<> (decimalTokens.subList(positionOfDecimalPoint +  1, decimalTokens.size())));

        }




        return toReturn;
    }

    private AstNumber parseNumber (ArrayList<Token> numberTokens) {

        AstNumber toReturn = new AstNumber();

        // take first Token from list and parse as digit without zero

        toReturn.astDigitWoz = parseDigitWoZ(numberTokens.get(0));

        if(numberTokens.size() == 1) {
            return toReturn;
        }

        numberTokens.stream()
                .skip(1)
                .forEach(token -> {
                    AstDigit temp = parseAstDigit(token);
                    toReturn.astDigits.add(temp);
                });

        return toReturn;

    }

    private AstDigitWoZ parseDigitWoZ (Token digitWoZ) {

        AstDigitWoZ toReturn = new AstDigitWoZ();

        int tokenContentAsString;

        try {
            tokenContentAsString = Integer.parseInt(digitWoZ.tokenString);
        } catch (NumberFormatException error) {
            System.out.println("Can't parse string in AstDigitWoZ");
            throw error;
        }


        toReturn.astDigitWozContent = tokenContentAsString;

        return toReturn;
    }

    private AstDigit parseAstDigit (Token digit) {

        AstDigit toReturn = new AstDigit();

        int tokenContentAsString;



        try {
            tokenContentAsString = Integer.parseInt(digit.tokenString);
        } catch (NumberFormatException error) {
            System.out.println("Can't parse string in AstDigit");
            throw error;
        }



        if(tokenContentAsString == 0) {
            toReturn.astDigit = 0;
            return toReturn;
        }

        toReturn.astDigitWoZ = parseDigitWoZ(digit);

        return toReturn;
    }


}
